from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from django.contrib.auth.backends import ModelBackend, AllowAllUsersModelBackend, RemoteUserBackend, AllowAllUsersRemoteUserBackend
from django.contrib.sites.models import Site

class GroupSiteModelBackendMixin():

    def _get_group_permissions(self, user_obj):
        '''
        Adds the SiteID for filtering Groups
        '''
        current_site = Site.objects.get_current()
        user_groups_field = get_user_model()._meta.get_field('groups')
        user_groups_query = 'group__%s' % user_groups_field.related_query_name()

        return Permission.objects.filter(**{user_groups_query: user_obj}).filter(group__permafrost_role__site=current_site)                   # TODO: Should it return Groups that do not have a Permafrost Role also?


class PermafrostModelBackend(GroupSiteModelBackendMixin, ModelBackend):
    '''
    Permafrost ModelBackend that takes into account SiteID when filtering on
    Group permissions via Permafrost Roles.
    '''
    pass


class PermafrostAllowAllUsersModelBackend(GroupSiteModelBackendMixin, AllowAllUsersModelBackend):
    '''
    Permafrost AllowAllUsersModelBackend that takes into account SiteID when filtering on
    Group permissions via Permafrost Roles.
    '''
    pass


class PermafrostRemoteUserBackend(GroupSiteModelBackendMixin, RemoteUserBackend):
    '''
    Permafrost RemoteUserBackend that takes into account SiteID when filtering on
    Group permissions via Permafrost Roles.
    '''
    pass


class PermafrostAllowAllUsersRemoteUserBackend(GroupSiteModelBackendMixin, AllowAllUsersRemoteUserBackend):
    '''
    Permafrost AllowAllUsersRemoteUserBackend that takes into account SiteID when filtering on
    Group permissions via Permafrost Roles.
    '''
    pass



# Generated by Django 3.0.3 on 2020-05-04 23:44

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('auth', '0011_update_proxy_permissions'),
        ('permafrost', '0009_permafrostrole_group_update'),
    ]

    operations = [
        migrations.AlterField(
            model_name='permafrostrole',
            name='group',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='permafrost_role', to='auth.Group', verbose_name='Group'),
        ),
    ]



# Generated by Django 3.1.3 on 2020-11-20 09:42

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('permafrost', '0016_merge_20200729_1716'),
    ]

    operations = [
        migrations.DeleteModel(
            name='PermafrostCategory',
        ),
    ]



# Generated by Django 3.0.3 on 2020-03-10 22:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('permafrost', '0003_auto_20200227_2356'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='permafrostrole',
            options={'permissions': (('add_user_to_role', 'Can Add Users to Role'),), 'verbose_name': 'Permafrost Role', 'verbose_name_plural': 'Permafrost Roles'},
        ),
        migrations.AddField(
            model_name='permafrostrole',
            name='deleted',
            field=models.BooleanField(default=False, help_text='Soft Delete the Role', verbose_name='Deleted'),
        ),
    ]



import json

from unittest import skipIf

from django.test import TestCase
from django.contrib.sites.models import Site
from django.contrib.auth import get_user_model
from django.db.utils import IntegrityError
from django.db import transaction
from django.contrib.auth.models import Group, Permission

try:
    from rest_framework.test import APIClient
    SKIP_DRF_TESTS = False
except ImportError:
    SKIP_DRF_TESTS = True

from permafrost.models import PermafrostRole


class PermafrostRoleModelTest(TestCase):

    fixtures = ['unit_test']

    def setUp(self):
        User = get_user_model()
        self.user = User.objects.create_user(username='john', email='jlennon@beatles.com', password='Passw0rd!')
        self.staffuser = User.objects.create_user(username='staffy', email='staffy@beatles.com', password='Passw0rd!')
        self.administrationuser = User.objects.create_user(username='adminy', email='adminy@beatles.com', password='Passw0rd!')

        self.site_1 = Site.objects.get(pk=1)
        self.site_2 = Site.objects.get(pk=2)

        self.perm_view_permafrostrole = Permission.objects.get_by_natural_key(*('view_permafrostrole', 'permafrost', 'permafrostrole')) 
        self.perm_change_permafrostrole = Permission.objects.get_by_natural_key(*('change_permafrostrole', 'permafrost', 'permafrostrole')) 
        self.perm_delete_permafrostrole = Permission.objects.get_by_natural_key(*('delete_permafrostrole', 'permafrost', 'permafrostrole')) 
        self.perm_add_logentry = Permission.objects.get_by_natural_key(*('add_logentry', 'admin', 'logentry')) 

    def test_role_rename_updates_group(self):
        '''
        Make sure renaming the PermafrostRole properly renames the Django Group model.
        '''
        role = PermafrostRole(name="Awesome Students", category="user")
        role.save()

        pk_check = role.group.pk
        self.assertEqual(role.group.name, "1_user_awesome-students")

        role.name = "OK Students"
        role.save()

        new_role_group = Group.objects.get(name=role.get_group_name())

        self.assertEqual(role.group.name, "1_user_ok-students")
        self.assertEqual(role.group.pk, pk_check)                   # Make sure a new group was not generated
        
    # User Roles

    def test_create_user_role(self):
        '''
        Test that creating a PermafrostRole creates a matching Group
        '''
        role = PermafrostRole(name="Bobs Super Group", category="user")
        role.save()
        role.users_add(self.user)
        perms = list(self.user.get_all_permissions())

        self.assertEqual(list(role.group.permissions.all()), [])            # Check the permissions on the group
        self.assertEqual(role.group.name, "1_user_bobs-super-group")        # Checks that the user is created
        self.assertEqual(perms, [])

    def test_add_optional_to_user_role(self):
        '''
        Test that the optional role can be added
        '''
        role = PermafrostRole(name="Bobs Super Group", category="user")
        role.save()
        role.permissions_add(self.perm_view_permafrostrole)
        role.users_add(self.user)
        perms = list(self.user.get_all_permissions())

        self.assertListEqual(list(role.group.permissions.all()), [self.perm_view_permafrostrole])            # Check the permissions on the group
        self.assertEqual(role.group.name, "1_user_bobs-super-group")        # Checks that the user is created
        self.assertListEqual(perms, ["permafrost.view_permafrostrole"])

    def test_add_not_allowed_to_user_role(self):
        '''
        Test that a permission that is not optional or required can be added
        '''
        role = PermafrostRole(name="Bobs Super Group", category="user")
        role.save()
        role.permissions_add(self.perm_delete_permafrostrole)
        role.users_add(self.user)
        perms = list(self.user.get_all_permissions())

        self.assertEqual(list(role.group.permissions.all()), [])            # Check the permissions on the group
        self.assertEqual(role.group.name, "1_user_bobs-super-group")        # Checks that the user is created
        self.assertListEqual(perms, [])

    def test_clear_permissions_on_user_role(self):
        '''
        Test that clearning permissions restores them to just the required.
        '''
        role = PermafrostRole(name="Bobs Super Group", category="user")
        role.save()
        role.permissions_add(self.perm_view_permafrostrole)
        role.permissions_clear()
        role.users_add(self.user)
        perms = list(self.user.get_all_permissions())

        self.assertEqual(list(role.group.permissions.all()), [])            # Check the permissions on the group
        self.assertEqual(role.group.name, "1_user_bobs-super-group")        # Checks that the user is created
        self.assertListEqual(perms, [])

    # Staff Roles

    def test_create_staff_role(self):
        role = PermafrostRole(name="Bobs Staff Group", category="staff")
        role.save()
        role.users_add(self.staffuser)      # Add user to the Group
        perms = list(self.staffuser.get_all_permissions())

        self.assertEqual([perm.name for perm in role.group.permissions.all()], ['Can view Role'])   # Make sure the required permission is present in the group
        self.assertEqual(role.group.name, "1_staff_bobs-staff-group")        # Checks that the user is created
        self.assertListEqual(perms, ['permafrost.view_permafrostrole'])

    def test_add_optional_to_staff_role(self):
        '''
        Test that the optional role can be added
        '''
        role = PermafrostRole(name="Bobs Staff Group", category="staff")
        role.save()
        role.permissions_add(self.perm_change_permafrostrole)
        role.users_add(self.staffuser)      # Add user to the Group
        perms = list(self.staffuser.get_all_permissions())
        perms.sort()

        self.assertListEqual(list(role.group.permissions.all()), [self.perm_change_permafrostrole, self.perm_view_permafrostrole])            # Check the permissions on the group
        self.assertEqual(role.group.name, "1_staff_bobs-staff-group")        # Checks that the user is created
        self.assertListEqual(perms, ['permafrost.change_permafrostrole', 'permafrost.view_permafrostrole'])

    def test_add_not_allowed_to_staff_role(self):
        '''
        Test that a permission that is not optional or required can be added
        '''
        role = PermafrostRole(name="Bobs Staff Group", category="staff")
        role.save()
        role.permissions_add(self.perm_delete_permafrostrole)
        role.users_add(self.staffuser)
        perms = list(self.staffuser.get_all_permissions())

        self.assertEqual([perm.name for perm in role.group.permissions.all()], ['Can view Role'])   # Make sure the required permission is present in the group
        self.assertEqual(role.group.name, "1_staff_bobs-staff-group")        # Checks that the user is created
        self.assertListEqual(perms, ['permafrost.view_permafrostrole'])

    def test_clear_permissions_on_staff_role(self):
        role = PermafrostRole(name="Bobs Staff Group", category="staff")
        role.save()
        role.permissions_add(self.perm_view_permafrostrole)
        role.permissions_clear()
        role.users_add(self.staffuser)      # Add user to the Group
        perms = list(self.staffuser.get_all_permissions())

        self.assertEqual([perm.name for perm in role.group.permissions.all()], ['Can view Role'])   # Make sure the required permission is present in the group
        self.assertEqual(role.group.name, "1_staff_bobs-staff-group")        # Checks that the user is created
        self.assertListEqual(perms, ['permafrost.view_permafrostrole'])

    # Administration Roles

    def test_create_administration_role(self):
        role = PermafrostRole(name="Bobs Administration Group", category="administration")
        role.save()
        role.users_add(self.administrationuser)      # Add user to the Group
        perms = list(self.administrationuser.get_all_permissions())
        perms.sort()

        self.assertListEqual([perm.name for perm in role.group.permissions.all()], ['Can add Role', 'Can change Role', 'Can view Role'])   # Make sure the required permission is present in the group
        self.assertEqual(role.group.name, "1_administration_bobs-administration-group")        # Checks that the user is created
        self.assertListEqual(perms, ['permafrost.add_permafrostrole', 'permafrost.change_permafrostrole', 'permafrost.view_permafrostrole'])

    def test_add_optional_to_administration_role(self):
        role = PermafrostRole(name="Bobs Administration Group", category="administration")
        role.save()
        role.permissions_add(self.perm_delete_permafrostrole)
        role.users_add(self.administrationuser)      # Add user to the Group
        perms = list(self.administrationuser.get_all_permissions())
        perms.sort()

        self.assertListEqual([perm.name for perm in role.group.permissions.all()], ['Can add Role', 'Can change Role', 'Can delete Role', 'Can view Role'])   # Make sure the required permission is present in the group
        self.assertEqual(role.group.name, "1_administration_bobs-administration-group")        # Checks that the user is created
        self.assertListEqual(perms, ['permafrost.add_permafrostrole', 'permafrost.change_permafrostrole', 'permafrost.delete_permafrostrole', 'permafrost.view_permafrostrole'])

    def test_add_not_allowed_to_administration_role(self):
        role = PermafrostRole(name="Bobs Administration Group", category="administration")
        role.save()
        role.permissions_add(self.perm_add_logentry)
        role.permissions_add(self.perm_delete_permafrostrole)
        role.users_add(self.administrationuser)      # Add user to the Group
        perms = list(self.administrationuser.get_all_permissions())
        perms.sort()

        self.assertListEqual([perm.name for perm in role.group.permissions.all()], ['Can add Role', 'Can change Role', 'Can delete Role', 'Can view Role'])   # Make sure the required permission is present in the group
        self.assertEqual(role.group.name, "1_administration_bobs-administration-group")        # Checks that the user is created
        self.assertListEqual(perms, ['permafrost.add_permafrostrole', 'permafrost.change_permafrostrole', 'permafrost.delete_permafrostrole', 'permafrost.view_permafrostrole'])

    def test_clear_permissions_on_administration_role(self):
        role = PermafrostRole(name="Bobs Administration Group", category="administration")
        role.save()
        role.permissions_add(self.perm_view_permafrostrole)
        role.permissions_clear()
        role.users_add(self.administrationuser)      # Add user to the Group
        perms = list(self.administrationuser.get_all_permissions())
        perms.sort()

        self.assertListEqual([perm.name for perm in role.group.permissions.all()], ['Can add Role', 'Can change Role', 'Can view Role'])   # Make sure the required permission is present in the group
        self.assertEqual(role.group.name, "1_administration_bobs-administration-group")        # Checks that the user is created
        self.assertListEqual(perms, ['permafrost.add_permafrostrole', 'permafrost.change_permafrostrole', 'permafrost.view_permafrostrole'])

    # Test Role Creation Rules

    def test_create_duplicate_role(self):
        '''
        Test that creating a PermafrostRole of the same name producers and error
        '''
        role_a = PermafrostRole(name="Bobs Super Group", site=self.site_1, category="user")
        role_a.save()

        role_c = PermafrostRole(name="Bobs Super Group", site=self.site_2, category="user")
        role_c.save()

        with self.assertRaises(IntegrityError):

            with transaction.atomic():
                role_b = PermafrostRole(name="Bobs Super Group", site=self.site_2, category="user")
                role_b.save()

            with transaction.atomic():
                role_d = PermafrostRole(name="Bobs Super Group", site=self.site_2, category="staff")
                role_d.save()

    # Test that deleting a PermafrostRole deletes the matching group

    def test_delete_role_deletes_group(self):
        role = PermafrostRole(name="Awesome Students", category="user")
        role.save()

        group = role.group
        group_name = group.name

        self.assertEqual(role.group.name, "1_user_awesome-students")

        role.delete()

        with self.assertRaises(Group.DoesNotExist):
            group = Group.objects.get(name=group_name)


# Don't run the following tests if DRF is not loaded
@skipIf(SKIP_DRF_TESTS, "Django Rest Framework not installed, skipping tests")
class PermafrostAPITest(TestCase):

    fixtures = ['unit_test']

    def setUp(self):
        User = get_user_model()
        self.user = User.objects.create_user(username='john', email='jlennon@beatles.com', password='Passw0rd!')
        self.staffuser = User.objects.create_user(username='staffy', email='staffy@beatles.com', password='Passw0rd!', is_active=True, is_staff=True, )
        self.adminuser = User.objects.create_user(username='adminy', email='adminy@beatles.com', password='Passw0rd!', is_active=True, is_staff=True, is_superuser=True)

        self.site_1 = Site.objects.get(pk=1)
        self.site_2 = Site.objects.get(pk=2)

        self.client = APIClient()

    def test_superuser_can_access_permissions_endpoint(self):
        '''
        Uses a user that has all the permissions.
        '''
        self.client.force_authenticate(user=self.adminuser)
        response = self.client.get('/permissions/', format='json')
        assert response.status_code == 200

    def test_can_not_access_permissions_endpoint(self):
        '''
        Uses a user that does not have the required permission
        '''
        self.client.force_authenticate(user=self.user)
        response = self.client.get('/permissions/', format='json')
        assert response.status_code == 403



from django.urls import path

from permafrost import views

app_name = 'permafrost'

urlpatterns = [
    path("roles/", views.PermafrostRoleListView.as_view(), name="role-list"),
    path("role/<slug:slug>/", views.PermafrostRoleDetailView.as_view(), name="role-detail"),
    path("role/<slug:slug>/update/", views.PermafrostRoleUpdateView.as_view(), name="role-update"),
    # path("role/add/", views.PermafrostRoleCreateView.as_view(), name="role-create"),
]





# Generated by Django 3.0.7 on 2020-06-06 00:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('permafrost', '0014_convert_category'),
    ]

    operations = [
        migrations.AlterField(
            model_name='permafrostrole',
            name='category',
            field=models.CharField(choices=[('administration', 'Administration'), ('staff', 'Staff'), ('user', 'User')], max_length=32, verbose_name='Category'),
        ),
    ]



# import sys
from django.conf import settings
from django.contrib.sites.models import Site
from django.db import models
from django.contrib.auth.models import Group, Permission
# from django.contrib.sites.shortcuts import get_current_site
from django.utils.translation import ugettext_lazy as _
from django.utils.text import slugify
from django.core.exceptions import ObjectDoesNotExist
from django.contrib.sites.managers import CurrentSiteManager
from django.db.models.signals import post_delete
from django.dispatch import receiver
from django.urls import reverse

from jsonfield import JSONField     # Using this instead of the PSQL one for portability


###############
# CHOICES
###############


try:
    CATEGORIES = getattr(settings, "PERMAFROST_CATEGORIES")
except AttributeError:
    print('''
    !!! Warning: PERMAFROST_CATEGORIES are not defined!

    They should look something like this and be defined in settings.py

    PERMAFROST_CATEGORIES = {
        'administration': {
            'label': _('Administration'),
            'level': 50,
            'optional': [
                {'label': _('Can delete Role'), 'permission': ('delete_permafrostrole', 'permafrost', 'permafrostrole') },
            ],
            'required': [
                {'label': _('Can add Role'), 'permission': ('add_permafrostrole', 'permafrost', 'permafrostrole') },
                {'label': _('Can change Role'), 'permission': ('change_permafrostrole', 'permafrost', 'permafrostrole') },
                {'label': _('Can view Role'), 'permission': ('view_permafrostrole', 'permafrost', 'permafrostrole') },
            ],
        },
        'staff': {
            'label': _('Staff'),
            'level': 30,
            'optional': [
                {'label': _('Can add Role'), 'permission': ('add_permafrostrole', 'permafrost', 'permafrostrole') },
                {'label': _('Can change Role'), 'permission': ('change_permafrostrole', 'permafrost', 'permafrostrole') },
                {'label': _('Can view Role'), 'permission': ('view_permafrostrole', 'permafrost', 'permafrostrole') },
            ],
            'required': [
                {'label': _('Can view Role'), 'permission': ('view_permafrostrole', 'permafrost', 'permafrostrole') },
            ],
        },
        'user': {
            'label': _('User'),
            'level': 1,
            'optional': [
                {'label': _('Can view Role'), 'permission': ('view_permafrostrole', 'permafrost', 'permafrostrole') },
            ],
            'required': [],
        },
    }

    See README.md for more information.
    ''')
    # sys.exit()
    raise


###############
# UTILITIES
###############

def get_current_site(*args, **kwargs):
    return settings.SITE_ID

###############
# MANAGERS
###############

class PermafrostRoleManager(models.Manager):
    '''
    Standard Django manager with natural key support added.
    '''
    def get_by_natural_key(self, slug, site):
        return self.get(slug=slug, site=site)


###############
# MODELS
###############

def get_choices():
    '''
    Creates a choice list based on the PERMAFROST_CATEGORIES settings.
    '''
    return [(cat_key, CATEGORIES[cat_key]['label']) for cat_key in CATEGORIES.keys()]


class PermafrostRole(models.Model):
    '''
    PermafrostRole is Client Defineable and "manages" a Django Group adding a
    user to this role adds them to the Django Group and automatically assignes
    them the permissions.

    The role is assigned to one of 3 categories to help group permission
    levels; 'administrator', 'staff' and 'user'.
    '''

    name = models.CharField(_("Name"), max_length=50)
    slug = models.SlugField(_("Slug"))
    category = models.CharField(_("Category"), max_length=32, choices=get_choices(), blank=False, null=False)                     # These should stay fixed to not trigger a potenital migration issue with changing choices
    site = models.ForeignKey(Site, on_delete=models.CASCADE, default=get_current_site)                      # This uses a callable so it will not trigger a migration with the projects it's included in
    locked = models.BooleanField(_("Locked"), default=False)                                                        # If this is locked, it can not be edited by the Client, used for System Default Roles
    deleted = models.BooleanField(_("Deleted"), default=False, help_text="Soft Delete the Role")
    group = models.ForeignKey(Group, verbose_name=_("Group"), on_delete=models.CASCADE, related_name="permafrost_role", blank=True, null=True)                  # NOTE: Need to make sure this is exported with natural key values as it can have a different PK on different servers

    objects = PermafrostRoleManager()
    on_site = CurrentSiteManager()

    class Meta:
        verbose_name = _("Permafrost Role")
        verbose_name_plural = _("Permafrost Roles")
        unique_together = (('name', 'site'),)

        permissions = (
            ("add_user_to_role", "Can Add Users to Role"),
            ("add_user_to_administration", "Can Add Users to the Administration Roles"),
        )

    def __str__(self):
        return self.name

    def natural_key(self):
        return (self.slug, self.site)

    def get_absolute_url(self):
        return reverse("permafrost:role-detail", kwargs={"slug": self.slug})
    
    def get_update_url(self):
        return reverse("permafrost:role-update", kwargs={"slug": self.slug})

    #-------------
    # Permissions

    def required_permissions(self):
        '''
        TODO: Read from the category and get the list of permissions 
        '''
        if 'required' in CATEGORIES[self.category]:
            return [Permission.objects.get_by_natural_key(*item['permission']) for item in CATEGORIES[self.category]['required']]
        return []

    def optional_permissions(self):
        '''
        TODO!!!
        '''
        if 'optional' in CATEGORIES[self.category]:
            return [Permission.objects.get_by_natural_key(*item['permission']) for item in CATEGORIES[self.category]['optional']]
        return []

    def all_perm_ids(self):
        req = [ perm.pk for perm in self.required_permissions() ]
        opt = [ perm.pk for perm in self.optional_permissions() ]
        return set(req + opt)

    def conform_group(self):
        '''
        TODO!!!
        Based on the list of permissions in the Category, make sure the group
        has the right set.  Make sure no permissions are outside of the
        optional and required and that all required permissions are added.

        Make sure the Group has the required permissions and all others are
        within the Optional permissions.
        '''
        self.permissions_set(self.group.permissions)        # This will check out to make sure all required permissions are present and optionals are allowed

    def get_group_name(self):
        '''
        Creates the standard name for the group
        '''
        return "{0}_{1}_{2}".format(self.site.pk, self.category, self.slug)
    
    def permissions(self):
        return self.group.permissions

    def permissions_add(self, *args):
        '''
        Add Django permission(s) to the attached group if the permission is in the allowed permissions
        '''
        id_check = self.all_perm_ids()
        for perm in args:
            if perm.pk in id_check:
                self.group.permissions.add(perm)

    def permissions_remove(self, *args):
        '''
        Remove Django permission(s) from the attached group if the permission is not in the list of required permissions
        '''
        id_check = [required.pk for required in self.required_permissions()]
        for perm in args:
            if perm.pk not in id_check:
                self.group.permissions.remove(perm)

    def permissions_set(self, permissions):
        '''
        This updates the group's Django permissions to only include what was passed in and passes the check against optional and required permissions.
        '''
        id_check = [required.pk for required in self.optional_permissions()]

        optional_perms = [perm for perm in permissions.all() if perm.pk in id_check]     # perms passed in that meet the optional filter check
        required_perms = self.required_permissions()

        # Set to values passed in that are in the optional list plus the required permissions.
        self.group.permissions.set(optional_perms + required_perms)

    def permissions_clear(self):          # TODO: Need to update
        '''
        Remove all Django permissions from the group except the required.
        '''
        if CATEGORIES[self.category]['required']:       # If there are any required permissions, set them
            self.group.permissions.set(self.required_permissions())
        else:                                           # Otherwise, clear it out completely
            self.group.permissions.clear()

    #-------------
    # Users

    def user_set(self):
        '''
        Wrapper around the group that returns a queryset of all the user 
        included in this role.
        '''
        return self.group.user_set.all()

    def users_add(self, *users):
        '''
        Pass in a User object to add to the PermafrostRole's Group
        '''
        self.group.user_set.add(*users)

    def users_remove(self, *users):
        '''
        Pass in a User object to remove from the PermafrostRole's Group
        '''
        self.group.user_set.remove(*users)

    def users_clear(self):
        '''
        Remove all users from the PermafrostRole's Group
        '''
        self.group.user_set.clear()

    #-------------
    # Save

    def save(self, *args, **kwargs):
        self.slug = slugify(self.name)
        group_name = self.get_group_name()

        if not self.pk:                                                             # if this is a new role, create the matching group
            self.group, created = Group.objects.get_or_create(name=group_name)      # Add the group if one named correctly alreay exists, otherwise create a new one.
        
        result = super().save(*args, **kwargs)

        if self.group.name != group_name:                                           # if the role is renamed after successful save, update the group's name
            self.group.name = group_name
            self.group.save()

        self.conform_group()                            # Apply after a successful save and Group creation (if needed)

        return result


@receiver(post_delete, sender=PermafrostRole, dispatch_uid='delete_matching_permafrost_role_group')
def delete_matching_group(sender, instance, using, **kwargs):
    instance.group.delete()



# Generated by Django 3.0.3 on 2020-03-19 00:22

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('permafrost', '0006_auto_20200313_2056'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='permafrostrole',
            name='group',
        ),
    ]



from pprint import pprint
import json

from django.contrib import admin
from django.shortcuts import render

from .models import PermafrostRole#, PermafrostCategory

################
# ADMIN ACTIONS
################


def perms_to_code(modeladmin, request, queryset):

    data = {}

    for role in queryset.all():
        
        if hasattr(role.category, 'name'):
            key = role.category.name.lower()
            label = role.category.name
        else:
            key = role.category
            label = role.get_category_display()
    
        if not key in data:       # First time though, assume everything available is required
            data[key] = {'req':list(role.group.permissions.all()), 'opt':[], 'label':label}
        else:
            new_perms = list(role.group.permissions.all())

            for perm in new_perms:
                if perm.pk not in [p.pk for p in data[key]['req']]:      # If the perm not shared with the previous, make it optional
                    data[key]['opt'].append(perm)

            new_req = []                            # Check previous required against the current
            for perm in data[key]['req']:
                if perm not in new_perms:               # If it's not shared, it's optional
                    data[key]['opt'].append(perm)
                else:                               # if it is shared, it's required
                    new_req.append(perm)

            data[key]['req'] = new_req

    # pprint(data, indent=4)

    key_order = list(data.keys())
    key_order.sort()

    result = ["from django.utils.translation import ugettext_lazy as _\n", "PERMAFROST_CATEGORIES = {"]

    for key in key_order:
        result.append("    '{}': {{".format(key))
        result.append("        'label': _('{}'),".format(data[key]['label']))
        
        if data[key]['opt']:
            result.append("        'optional': [")
            for perm in data[key]['opt']:
                result.append("            {{'label': _('{}'), 'permission': {} }},".format(perm.name, perm.natural_key()) )
            result.append("        ],")
        else:
            result.append("        'optional': [],")

        if data[key]['req']:
            result.append("        'required': [")
            for perm in data[key]['req']:
                result.append("            {{'label': _('{}'), 'permission': {} }},".format(perm.name, perm.natural_key()) )
            result.append("        ],")
        else:
            result.append("        'required': [],")

        result.append("    },")
    result.append("}\n")


        # data[key]['optional'] = [{'label':perm.name, 'permission':perm.natural_key()} for perm in data[key].pop('opt')]
        # data[key]['required'] = [{'label':perm.name, 'permission':perm.natural_key()} for perm in data[key].pop('req')]

    return render(request, 'permafrost/admin/perms_to_code.html', context={'data':"\n".join(result)})
    # return render(request, 'permafrost/admin/perms_to_code.html', context={'json_data':"PERMAFROST_CATEGORIES = {}".format(json.dumps(data, indent=4))})


perms_to_code.short_description = "Convert Model Permissions to Code"


def create_missing_groups(modeladmin, request, queryset):

    for item in queryset.all():
        # Make sure group exists and create if not after model is saved.
        item.get_group()


create_missing_groups.short_description = "Create a Django Group if missing"


###############
# MODEL ADMINS
###############

class PermafrostRoleAdmin(admin.ModelAdmin):
    readonly_fields = ('slug',)
    list_display = ('name', 'category', 'group', 'site')
    ordering = ('name',)
    readonly_fields=('slug',)
    actions = [create_missing_groups, perms_to_code]


# class PermafrostCategoryAdmin(admin.ModelAdmin):
#     readonly_fields = ('slug',)
#     list_display = ('name',)
#     ordering = ('name',)
#     readonly_fields=('slug',)

###############
# REGISTRATION
###############

admin.site.register(PermafrostRole, PermafrostRoleAdmin)
# admin.site.register(PermafrostCategory, PermafrostCategoryAdmin)



# Generated by Django 3.0.7 on 2020-06-05 23:15

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('permafrost', '0011_auto_20200506_2128'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='permafrostrole',
            name='category',
        ),
    ]



# Generated by Django 3.0.7 on 2020-07-29 17:16

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('permafrost', '0012_auto_20200729_1710'),
        ('permafrost', '0015_auto_20200606_0042'),
    ]

    operations = [
    ]



import logging

from django.contrib.auth.decorators import login_required
from django.contrib.auth.mixins import PermissionRequiredMixin
from django.http import HttpResponse, HttpResponseForbidden
from django.shortcuts import render
from django.core.exceptions import PermissionDenied
from django.views.generic import TemplateView, ListView, DetailView, UpdateView, DeleteView
from unittest import skipIf

from .models import PermafrostRole

#--------------
# UTILITIES
#--------------

def get_client_ip(request):
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip


#--------------
# MIXIN VIEWS
#--------------

class PermafrostMixin(PermissionRequiredMixin):
    '''
    This is a simple mixin that extend the built in PermissionRequiredMixin 
    and lets a developer specify perms required by a user for a particular 
    http method.  For example, a user with a 'get' pemission might not be
    given permissions to reach a 'post' endpoint in the view.
    
    If they don't have the required permissions, then the user
    is rejected.

    Permissions can be set per HTTP method, by appending the lowercase
    method name to 'permission_required_' and providing a set of permissions.

    permission_required = ('sites.add_site',)
    permission_required_get = ()
    permission_required_post = ()
    '''

    def get_permission_required(self):
        """
        Override this method to override the permission_required attribute.
        Must return an iterable.
        """
        perms = super().get_permission_required()

        method_perms = getattr(self, "permission_required_" + self.request.method.lower(), set() )       # Extended Perms per method

        if isinstance(method_perms, str):
            method_perms = (method_perms,)

        return set(list(perms) + list(method_perms))


class PermafrostLogMixin(object):
    '''
    A mixin that lets you define a logger in which to write failed permission attempts to.
    '''

    permission_logger = None

    def handle_no_permission(self):
        
        if self.permission_logger is None:  # TODO Make this assume a default logger called "permafrost"
            raise ImproperlyConfigured(
                '{0} is missing the permission_logger attribute. Define {0}.permission_logger'.format(self.__class__.__name__)
            )

        logger = logging.getLogger(self.permission_logger)

        user_ip = get_client_ip(self.request)
        user_perms = list(self.request.user.get_all_permissions())
        view_perms = list(self.get_permission_required())

        logger.info("Failed-Permission-Check:403:{0}:{1}:{2}:{3}:{4}:{5}:{6}".format(                   # Should be replaced with a Formater
                                user_ip, self.request.user.username, self.request.user.pk, 
                                self.request.method, self.request.path, ','.join(user_perms),
                                ','.join(view_perms) )
                            )

        super().handle_no_permission()


# Create Permission Group
class PermafrostRoleCreateView(ListView):
    model = PermafrostRole

# List Permission Groups
class PermafrostRoleListView(ListView):
    model = PermafrostRole

# Detail Permission Groups
class PermafrostRoleDetailView(DetailView):
    model = PermafrostRole

# Update Permission Group
class PermafrostRoleUpdateView(UpdateView):
    model = PermafrostRole
    fields = ['name', 'locked', 'deleted']

# Delete Permission Groups
class PermafrostRoleDeleteView(DeleteView):
    model = PermafrostRole


# Future Views

# TODO: Role User List (For easier pagination) & bulk editing

# TODO: User Roles on a given Site

# TODO: Roles with a given permission



# Generated by Django 3.0.7 on 2020-06-05 23:33

from django.db import migrations

def set_missing_category(apps, schema_editor):
    PermafrostRole = apps.get_model('permafrost', 'PermafrostRole')
    for role in PermafrostRole.objects.all():
        if not role.category and role.group:
            role.category = role.group.name.split("_")[1]
            role.save()


class Migration(migrations.Migration):

    dependencies = [
        ('permafrost', '0013_permafrostrole_category'),
    ]

    operations = [
        migrations.RunPython(set_missing_category, reverse_code=migrations.RunPython.noop),
    ]



# Generated by Django 3.0.7 on 2020-07-29 17:10

import django.contrib.sites.managers
from django.db import migrations
import django.db.models.manager


class Migration(migrations.Migration):

    dependencies = [
        ('permafrost', '0011_auto_20200506_2128'),
    ]

    operations = [
        migrations.AlterModelManagers(
            name='permafrostrole',
            managers=[
                ('objects', django.db.models.manager.Manager()),
                ('on_site', django.contrib.sites.managers.CurrentSiteManager()),
            ],
        ),
    ]



# Generated by Django 3.0.4 on 2020-03-11 17:01

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('permafrost', '0004_auto_20200310_2228'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='permafrostrole',
            options={'permissions': (('add_user_to_role', 'Can Add Users to Role'), ('add_user_to_administration', 'Can Add Users to the Administration Roles')), 'verbose_name': 'Permafrost Role', 'verbose_name_plural': 'Permafrost Roles'},
        ),
    ]
